import asyncio
from copy import deepcopy
from datetime import datetime
import inspect
from uuid import uuid4

from tukio.broker import get_broker, EXEC_TOPIC
from tukio.event import EventSource, Event
from tukio.utils import FutureState

from .task import TaskRegistry, TaskExecState


class TukioTask(asyncio.Task):

    """
    A simple subclass of `asyncio.Task()` to add an execution ID and optionally
    bind a task holder class.
    """

    def __init__(self, coro, *, loop=None):
        super().__init__(coro, loop=loop)
        self.holder = inspect.getcoroutinelocals(coro).get('self')
        try:
            self.uid = self.holder.uid
        except AttributeError:
            self.uid = str(uuid4())
        self._broker = get_broker(self._loop)
        self._in_progress = False
        self._template = None
        self._workflow = None
        self._source = None
        self._start = None
        self._end = None
        self._inputs = None
        self._outputs = None
        self._queue = asyncio.Queue(loop=self._loop)
        if self.holder:
            self.holder.queue = self._queue

    @property
    def inputs(self):
        return self._inputs

    @inputs.setter
    def inputs(self, data):
        try:
            if isinstance(data, Event):
                self._inputs = deepcopy(data.data)
            else:
                self._inputs = deepcopy(data)
        except TypeError:
            self._inputs = {"error": "Cannot display task result"}

    @property
    def template(self):
        return self._template

    @property
    def workflow(self):
        return self._workflow

    @property
    def event_source(self):
        return self._source

    @property
    def queue(self):
        return self._queue

    def as_dict(self):
        """
        Returns the execution informations of this task.
        """
        return {
            'id': self.uid,
            'start': self._start,
            'end': self._end,
            'state': FutureState.get(self).value,
            'inputs': self._inputs,
            'outputs': self._outputs
        }

    async def data_received(self, event):
        """
        A handler that puts events into the task's own receiving queue. This
        handler shall be registered in the data broker so that any tukio task
        can receive and process events during execution.
        """
        await self._queue.put(event)

    def in_progress(self):
        """
        Returns True if the task execution started, else returns False.
        """
        return self._in_progress

    def set_result(self, result):
        """
        Wrapper around `Future.set_result()` to automatically dispatch a
        `TaskExecState.end` event.
        """
        super().set_result(result)
        try:
            if isinstance(result, Event):
                self._outputs = deepcopy(result.data)
            else:
                self._outputs = deepcopy(result)
        except TypeError:
            self._outputs = {"error": "Cannot display task result"}

        self._end = datetime.utcnow()
        data = {'type': TaskExecState.end.value, 'content': self._outputs}
        self._broker.dispatch(data=data, topic=EXEC_TOPIC, source=self._source)

    def set_exception(self, exception):
        """
        Wrapper around `Future.set_exception()` to automatically dispatch a
        `TaskExecState.error` event.
        """
        super().set_exception(exception)
        self._end = datetime.utcnow()
        data = {'type': TaskExecState.error.value, 'content': exception}
        self._broker.dispatch(data=data, topic=EXEC_TOPIC, source=self._source)

    def _step(self, exc=None):
        """
        Wrapper around `Task._step()` to automatically dispatch a
        `TaskExecState.begin` event.
        """
        if not self._in_progress:
            self._start = datetime.utcnow()
            source = {'task_exec_id': self.uid}
            if self._template:
                source['task_template_id'] = self._template.uid
            if self._workflow:
                source['workflow_template_id'] = self._workflow.template.uid
                source['workflow_exec_id'] = self._workflow.uid
            self._source = EventSource(**source)
            self._in_progress = True
            data = {
                'type': TaskExecState.begin.value,
                'content': self._inputs
            }
            self._broker.dispatch(data, topic=EXEC_TOPIC, source=self._source)
        super()._step(exc)


def tukio_factory(loop, coro):
    """
    A task factory for asyncio that creates `TukioTask()` instances for all
    coroutines registered as Tukio tasks and default `asyncio.Task()` instances
    for all others.
    """
    try:
        # Trigger exception if not valid
        TaskRegistry.codes()[coro.cr_code]
        task = TukioTask(coro, loop=loop)
    except (KeyError, AttributeError):
        # When the coroutine is not a registered Tukio task or when `coro` is a
        # simple generator (e.g. upon calling `asyncio.wait()`)
        task = asyncio.Task(coro, loop=loop)
    return task



"""
That module assumes that a Tukio task runs as an `asyncio.Task()` instance.

It provides a registry `TaskRegistry` to register classes that define
coroutines as well as standard standalone coroutines as Tukio tasks.

It also provides a function `new_task()` that shall be used to schedule the
execution of Tukio tasks in an asyncio event loop.
"""
import asyncio
import logging
import inspect
from enum import Enum

from tukio.broker import get_broker, EXEC_TOPIC


log = logging.getLogger(__name__)


class UnknownTaskName(Exception):
    pass


# Inspired from https://github.com/faif/python-patterns/blob/master/registry.py
class TaskRegistry:

    """
    This class is designed to be a registry of all coroutines and classes that
    implement a Tukio task. Coroutines and classes are registered by name.
    Each record of the registry is a tuple (class, coro_func). If the
    registered implementation is a simple coroutine, class=None.
    """

    _registry = dict()
    _codes = dict()

    @classmethod
    def register(cls, coro_or_cls, task_name, coro_name=None):
        # A task name can be registered only once!
        if task_name in cls._registry:
            klass, coro_fn = cls._registry[task_name]
            raise ValueError("task name '{}' already registered with "
                             "({}, {})".format(task_name, klass, coro_fn))

        if inspect.isclass(coro_or_cls):
            klass, coro_fn = coro_or_cls, getattr(coro_or_cls, coro_name)
        else:
            klass, coro_fn = None, coro_or_cls
        if not asyncio.iscoroutinefunction(coro_fn):
            raise TypeError("{} is not a coroutine function".format(coro_fn))
        cls._registry[task_name] = klass, coro_fn
        cls._codes[coro_fn.__code__] = task_name

    @classmethod
    def get(cls, task_name):
        try:
            return cls._registry[task_name]
        except KeyError as exc:
            raise UnknownTaskName from exc

    @classmethod
    def all(cls):
        return cls._registry

    @classmethod
    def codes(cls):
        return cls._codes


def register(task_name, coro_name=None):
    """
    A decorator to register a standard coroutine or a task holder class as a
    Tukio task implementation.
    This decorator adds the class attribute `TASK_NAME` to task holder classes.
    """
    def decorator(coro_or_cls):
        TaskRegistry.register(coro_or_cls, task_name, coro_name)
        if inspect.isclass(coro_or_cls):
            coro_or_cls.TASK_NAME = task_name
        return coro_or_cls
    return decorator


def new_task(task_name, *, data=None, config=None, loop=None):
    """
    Schedules the execution of the coroutine registered as `task_name` (either
    defined in a task holder class or not) in the loop and returns an instance
    of `asyncio.Task()` (or a subclass of it).
    """
    klass, coro_fn = TaskRegistry.get(task_name)
    if klass:
        task_holder = klass(config)
        coro = coro_fn(task_holder, data)
    else:
        coro = coro_fn(data)
    return asyncio.ensure_future(coro, loop=loop)


class TaskExecState(Enum):

    begin = 'task-begin'
    end = 'task-end'
    error = 'task-error'
    progress = 'task-progress'


def dispatch_progress(data, broker=None):
    """
    A shorthand to dispatch task progress events in the 'EXEC_TOPIC' from any
    tukio task.
    """
    broker = broker or get_broker()
    task = asyncio.Task.current_task()
    event_data = {'type': TaskExecState.progress.value}
    event_data['content'] = data
    broker.dispatch(event_data, topic=EXEC_TOPIC, source=task.event_source)



from .join import JoinTask
from .task import TaskRegistry, UnknownTaskName, register, new_task, dispatch_progress
from .factory import TukioTask, tukio_factory
from .holder import TaskHolder
from .template import TaskTemplate



"""
Any data received by the engine is wrapped into an event and passed to
workflows and tasks.
"""
import collections


_BaseEvent = collections.namedtuple(
    '_BaseEvent', ('data', 'topic', 'source'))
# Init default values to None, except 'data' which is always mandatory
_BaseEvent.__new__.__defaults__ = (None,) * (len(_BaseEvent._fields) - 1)


class Event(_BaseEvent):

    """
    An event is a structure that gathers data and optionally the topic it shall
    be dispatched to and the source of the event.
    The registered coroutine of each task receives an `Event` object as
    argument. The same way, `data_received` (only for task holders) is also
    called with an `Event` object as argument.
    """

    __slots__ = ()


_BaseEventSource = collections.namedtuple(
    '_BaseEventSource', ('workflow_template_id', 'workflow_exec_id',
                         'task_template_id', 'task_exec_id'))
# Init default values to None
_BaseEventSource.__new__.__defaults__ = (None,) * len(_BaseEventSource._fields)


class EventSource(_BaseEventSource):

    """
    An event source is a structure that gathers everything to know where an
    event comes from and/or the topics it will be dispatched to.
    """

    __slots__ = ()
